from flymyai.core.clients.AsyncClient import BaseAsyncClient
from flymyai.core.clients.SyncClient import BaseSyncClient


class FlyMyAI(BaseSyncClient): ...


class AsyncFlyMyAI(BaseAsyncClient): ...
